'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = build;

require('babel-polyfill');

/**
 * # Closure Eliminator
 */

var $boundArrowFunction = Symbol('boundArrowFunction');
var $usedEval = Symbol('usedEval');

function _ref2(_ref) {
  var node = _ref.node;
  return node._generated || node._compact;
}

function _ref(path) {
  return path.type === 'ArrowFunctionExpression';
}

function _ref3(parentArrow) {
  parentArrow.node[$boundArrowFunction] = true;
}

function _enter(path) {
  path.getAncestry().filter(_ref).forEach(_ref3);
}

function _ref4(path) {
  return path.isFunction();
}

function _ref5(parentArrow) {
  parentArrow.node[$usedEval] = true;
}

function _enter2(path) {
  if (path.node.name === 'eval' && path.parentPath.type === 'CallExpression') {
    path.getAncestry().filter(_ref4).forEach(_ref5);
  }
}

function _getAllParentScopes(scope) {
  var scopes = [];
  while (scope = scope.parent) {
    scopes.push(scope);
  }
  return scopes;
}

function _getAttachmentPositio(bestParent, prevPath) {
  var possibleSiblings = void 0;
  if (bestParent.isFunction()) {
    possibleSiblings = bestParent.get('body.body');
  } else if (bestParent.isProgram() || bestParent.isBlockStatement()) {
    possibleSiblings = bestParent.get('body');
  }
  var prevParents = prevPath.getAncestry();
  return possibleSiblings.find(function (x) {
    return prevParents.indexOf(x) !== -1;
  });
}

function build(babel) {
  var t = babel.types;


  return {
    visitor: {
      Function: {
        exit: function exit(path) {
          var node = path.node;

          if (path.isClassMethod() || path.isObjectMethod() || node[$boundArrowFunction] || node[$usedEval]) {
            return;
          }
          if (path.findParent(_ref2)) {
            path.skip();
            return;
          }
          var bestParentScope = getHighestCompatibleHoistedScope(path);
          if (bestParentScope !== path.scope.parent) {
            var attachPath = _getAttachmentPositio(bestParentScope.path, path);
            moveToNewPosition(path, attachPath);
          }
        }
      },
      ThisExpression: {
        enter: _enter
      },
      Identifier: {
        enter: _enter2
      }
    }
  };

  function getHighestCompatibleHoistedScope(path) {
    var parentScopes = _getAllParentScopes(path.scope),
        parentBindings = path.scope.parent.getAllBindings();

    function _ref6(subPath) {
      return subPath.getAncestry().indexOf(path) !== -1;
    }

    for (var id in parentBindings) {
      var hasUsageOfBinding = [].concat(parentBindings[id].referencePaths).concat(parentBindings[id].constantViolations).some(_ref6);
      if (hasUsageOfBinding) {
        var idx = parentScopes.indexOf(parentBindings[id].scope);
        if (idx !== -1) {
          parentScopes.splice(idx + 1, Infinity);
        }
      }
    }
    return parentScopes.pop();
  }function moveToNewPosition(path, attachPath) {
    var node = path.node,
        scope = path.scope,
        newScope = attachPath.parentPath.scope;

    if (node.type === 'FunctionDeclaration') {
      var uid = newScope.generateUidIdentifierBasedOnNode(node.id);
      scope.rename(node.id.name, uid.name);
      scope.moveBindingTo(uid.name, newScope);
      node._hoisted = true;
      attachPath.insertBefore([node]);
      path.remove();
    } else {
      var _uid = path.parentPath.scope.generateUidIdentifierBasedOnNode(node.id),
          replacement = t.functionDeclaration(_uid, node.params, normalizeFunctionBody(node.body));
      replacement.loc = node.loc;
      replacement.generator = node.generator;
      replacement.async = node.async;
      replacement._hoisted = true;
      attachPath.insertBefore([replacement]);
      path.replaceWith(t.identifier(_uid.name));
      newScope.crawl();
    }
  }

  /**
   * Normalize a function body so that it is always a BlockStatement.
   */
  function normalizeFunctionBody(node) {
    if (node.type !== 'BlockStatement') {
      return t.blockStatement([t.returnStatement(node)]);
    } else {
      return node;
    }
  }
}